/*
 *  Video stacking program by Panu Lahtinen, pnuu+finsprite@iki.fi
 */

#include "finsprite.h"

void create_md_zones(int img_width, int img_height){

  unsigned char *mask = NULL;
  int mask_width = 0;
  int mask_height = 0;

  int i = 0;

  /* Read mask, if available */
  if(md_mask_fname != NULL) {
    /* read */
    mask = read_mask(md_mask_fname, &mask_width, &mask_height);
  }

  if(mask_width != img_width || mask_height != img_height){
    if(mask != NULL){ free(mask); }
    /* Default to allow all pixels to be used in motion detection */
    mask = (unsigned char*)malloc(img_width*img_height*sizeof(unsigned char));
    for(i=0;i<img_width*img_height;i++){ mask[i] = 1; }
    //    memset(mask,1,img_width*img_height);
    mask_width = mask_width;
    mask_height = img_height;
  }

  /* Setup zone sums */
  md_zone_sums = (int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(int));
  md_zone_sums_new = (int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(int));
  for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
    md_zone_sums[i] = INT_MAX;
    md_zone_sums_new[i] = 0;
  }

  /* Number of valid pixels in each motion detection zones */
  md_zone_num_pix = (unsigned int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(unsigned int));
  memset(md_zone_num_pix,0,md_num_x_zones*md_num_y_zones*sizeof(unsigned int));

  /* Setup zone reference */
  md_zone_reference = (int*)malloc(img_width*img_height*sizeof(int));

  /* Pixels per zone in each direction */
  float pix_per_x_zone = (img_width/(float)md_num_x_zones);
  float pix_per_y_zone = (img_height/(float)md_num_y_zones);

  int x_zone = 0;
  int y_zone = 0;
  int row = -1;

  /* Form the reference vector that shows to which zone each pixel belongs to */
  for(i=0; i < img_width*img_height; i++){
    if(i%img_width == 0){
      row++;
    }

    if(mask[i] != 0){
      x_zone = (int)((i-row*img_width)/pix_per_x_zone);
      y_zone = (int)(row/pix_per_y_zone);
      md_zone_reference[i] = x_zone + y_zone * md_num_x_zones;
      md_zone_num_pix[md_zone_reference[i]]++;
    } else {
      md_zone_reference[i] = -1;
      printf("hop\n");
    }
           
    //if(i == 414720-1){ printf("%d,%d,%d,%d,%f,%f,%d\n",x_zone,y_zone,md_zone_reference[i],row,pix_per_x_zone,pix_per_y_zone,i); }

  }
  if(mask != NULL){
    free(mask);
  }
  printf("Zones ready\n");
}

static void process_image(const void * p, time_t timestamp, long usec){

  static int packed_value;
  int i = 0, j = 0;
  int Y0, Y1, Y2, Y3;

  int x=0, y=0;
  int frame_ok = 0;
  unsigned long min = ULONG_MAX;

  char *fname = NULL;

  if(use_trig){

    if(md_frames[md_i].img_buffer != NULL){
      free(md_frames[md_i].img_buffer);
      md_frames[md_i].img_buffer = NULL;
    }
    if(md_frames[md_i].img_buffer == NULL){
      md_frames[md_i].img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*md_frames[md_i].img_buffer));
      md_frames[md_i].width = fmt.fmt.pix.width;
      md_frames[md_i].height = fmt.fmt.pix.height;
    }

    /* Form the motion detection zones */
    if(md_zone_reference == NULL){
      create_md_zones(fmt.fmt.pix.width, fmt.fmt.pix.height);
    } else {
      /* Or if already done, update comparison values and reset for new image
      */
      for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
	md_zone_sums[i] = md_zone_sums_new[i];
	md_zone_sums_new[i] = 0;
      }
    }

    /* Update the timestamp */
    md_frames[md_i].timestamp = timestamp;
    md_frames[md_i].usec = usec;
  }

  if(use_max || use_dif){
    if(peak_hold.img_buffer == NULL){
      peak_hold.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*peak_hold.img_buffer));
      peak_hold.width = fmt.fmt.pix.width;
      peak_hold.height = fmt.fmt.pix.height;
      peak_hold.timestamp = timestamp;
      peak_hold.usec = usec;
      stack_timestamp = timestamp;
    }
  }

  if(use_min || use_dif){
    if(min_hold.img_buffer == NULL){
      min_hold.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*min_hold.img_buffer));
      for(i=0;i<fmt.fmt.pix.width*fmt.fmt.pix.height;i++){
	min_hold.img_buffer[i] = 255;
      }
      min_hold.width = fmt.fmt.pix.width;
      min_hold.height = fmt.fmt.pix.height;
      min_hold.timestamp = timestamp;
      min_hold.usec = usec;
      stack_timestamp = timestamp;
    }
  }

  if(use_ave){
    if(ave_stack.img_buffer == NULL){
      ave_stack.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height, sizeof(*ave_stack.img_buffer));
      ave_stack.width = fmt.fmt.pix.width;
      ave_stack.height = fmt.fmt.pix.height;
      ave_stack.timestamp = timestamp;
      ave_stack.usec = usec;
      ave_stack.max_val = 0;
      ave_stack.min_val = ULONG_MAX;
      stack_timestamp = timestamp;
    }
  }

  i = 0;
  while(j+3 < (fmt.fmt.pix.width * fmt.fmt.pix.height)){
    packed_value = *((int*)p+i);

    switch(fmt.fmt.pix.pixelformat){
    case V4L2_PIX_FMT_GREY:
      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 4) & 0xFF);
      Y1 = (char)((packed_value >> 8) & 0xFF);
      Y3 = (char)((packed_value >> 12) & 0xFF);

      break;
    case V4L2_PIX_FMT_UYVY:
    case V4L2_PIX_FMT_YUYV:

      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 8) & 0xFF);
      Y1 = (char)((packed_value >> 16) & 0xFF);
      Y3 = (char)((packed_value >> 24) & 0xFF);

      break;
    default:
      break;
    }

    // Strip sign values after shift (i.e. unsigned shift)
    Y0 = Y0 & 0xFF;
    Y1 = Y1 & 0xFF;
    Y2 = Y2 & 0xFF;
    Y3 = Y3 & 0xFF;

    x = (i+1) % (fmt.fmt.pix.width/2) - 1;
    y = (i+1) / (fmt.fmt.pix.width/2);
    j = x*2 + y*fmt.fmt.pix.width;

    if(j+3 > fmt.fmt.pix.width * fmt.fmt.pix.height){ break; }

    if(use_trig){
      md_frames[md_i].img_buffer[j+0] = Y0;
      md_frames[md_i].img_buffer[j+1] = Y1;
      md_frames[md_i].img_buffer[j+2] = Y2;
      md_frames[md_i].img_buffer[j+3] = Y3;
    }

    if(use_max || use_dif){
      if(Y0 > peak_hold.img_buffer[j+0]){ peak_hold.img_buffer[j+0] = Y0; }
      if(Y1 > peak_hold.img_buffer[j+1]){ peak_hold.img_buffer[j+1] = Y1; }
      if(Y2 > peak_hold.img_buffer[j+2]){ peak_hold.img_buffer[j+2] = Y2; }
      if(Y3 > peak_hold.img_buffer[j+3]){ peak_hold.img_buffer[j+3] = Y3; }
    }

    if(use_min || use_dif){
      if(Y0 < min_hold.img_buffer[j+0]){ min_hold.img_buffer[j+0] = Y0; }
      if(Y1 < min_hold.img_buffer[j+1]){ min_hold.img_buffer[j+1] = Y1; }
      if(Y2 < min_hold.img_buffer[j+2]){ min_hold.img_buffer[j+2] = Y2; }
      if(Y3 < min_hold.img_buffer[j+3]){ min_hold.img_buffer[j+3] = Y3; }
    }

    if(use_ave){
      ave_stack.img_buffer[j+0] += Y0;
      ave_stack.img_buffer[j+1] += Y1;
      ave_stack.img_buffer[j+2] += Y2;
      ave_stack.img_buffer[j+3] += Y3;

      if(ave_stack.img_buffer[j+0] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+0]; }
      if(ave_stack.img_buffer[j+1] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+1]; }
      if(ave_stack.img_buffer[j+2] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+2]; }
      if(ave_stack.img_buffer[j+3] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+3]; }
      
      if(ave_stack.img_buffer[j+0] < min){ min = ave_stack.img_buffer[j+0]; }
      if(ave_stack.img_buffer[j+1] < min){ min = ave_stack.img_buffer[j+1]; }
      if(ave_stack.img_buffer[j+2] < min){ min = ave_stack.img_buffer[j+2]; }
      if(ave_stack.img_buffer[j+3] < min){ min = ave_stack.img_buffer[j+3]; }
    }

    if(Y0 > 0 && Y1 > 0 && Y2 > 0 && Y3 > 0){
      frame_ok = 1;
    }

    if(use_trig && md_detection == -1){
      md_zone_sums_new[md_zone_reference[j+0]] += Y0;
      md_zone_sums_new[md_zone_reference[j+1]] += Y1;
      md_zone_sums_new[md_zone_reference[j+2]] += Y2;
      md_zone_sums_new[md_zone_reference[j+3]] += Y3;
    }

    i++;
  }

  /* Save ring memory to images on disc */
  //  if(md_images_since_detection+1 > img_in_md_buffer/2){
  if(100*md_images_since_detection/img_in_md_buffer >= md_percentage){
    for(i=0; i<img_in_md_buffer; i++){
      fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      make_fname(fname,md_frames[i].timestamp, md_frames[i].usec, TRIG);
      if(!md_test){
	printf("[%d] %s saved from ring memory\n",i,fname);
	writeImage(fname, md_frames[i].width, md_frames[i].height, md_frames[i].img_buffer, "Finsprite motion detection at work");
      } else {
	printf("[%d] %s read from ring memory\n",i,fname);
      }
      free(fname);
    }
    md_images_since_detection = -1;
    md_detection = -1;
    for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
      md_zone_sums[i] = INT_MAX;
      md_zone_sums_new[i] = INT_MAX;
    }
      //    memset(md_zone_sums,INT_MAX,md_num_x_zones*md_num_y_zones*sizeof(int));
      //    memset(md_zone_sums_new,INT_MAX,md_num_x_zones*md_num_y_zones*sizeof(int));
    md_last_detection = timestamp+DELAY_BETWEEN_MD;
  }

  /* Detect rapid changes in illumination (motion detection) */
  if(previous_frame >= 0 && use_trig && md_detection == -1 && 
     timestamp-md_last_detection > DELAY_BETWEEN_MD &&
     timestamp-start_time > 2*DELAY_BETWEEN_MD){
    for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
      /* Use only zones with pixels */
      if(md_zone_num_pix[i] > 0){
	/* Trigger if the new image zone was brighter */
	if((100.0*md_zone_sums_new[i]/md_zone_num_pix[i] -
	    100.0*md_zone_sums[i]/md_zone_num_pix[i])
	   > md_threshold){
	  
	  char *tmp = make_time_string(timestamp,usec,TRIG);
	  printf("Zone %d triggered: %f (%d) at %s\n",i,100.0*md_zone_sums_new[i]/md_zone_num_pix[i]-100.0*md_zone_sums[i]/md_zone_num_pix[i],md_threshold,tmp);
	  free(tmp);
	  md_detection = md_i;
	  md_images_since_detection = 0;
	  md_last_detection = timestamp;
	  break;
	}
      }
      //printf("zone %d has %d pixels\n",i,md_zone_num_pix[i]);
    }
  }

  if(md_detection != -1){
    md_images_since_detection++;
  }
  
  previous_frame = md_i;
  // Increment index where to store image
  md_i++;
  // If we are using the last slot of the image buffer, start from the zero
  if(md_i >= img_in_md_buffer){
    md_i = 0;
  }

  if(frame_ok && use_ave){
    ave_stack.num_imgs++;

    if(min > ave_stack.min_val){
      ave_stack.min_val = min;
    }
  }

  if(stack_length < difftime(timestamp,stack_timestamp)){

    if(use_ave){
      save_ave_stack();
    }

    if(use_dif){
      save_dif_stack();
    }

    if(use_max){
      fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      make_fname(fname, peak_hold.timestamp, peak_hold.usec, PEAK_HOLD);
      printf("Saving %s\n",fname);
      //stretch(peak_hold.img_buffer, peak_hold.width*peak_hold.height, 0.01, 0.99);
      writeImage(fname, peak_hold.width, peak_hold.height, peak_hold.img_buffer, "Finsprite peak-hold stack");
      free(fname);
    }

    if(use_max || use_dif){
      free(peak_hold.img_buffer); peak_hold.img_buffer = NULL;
    }

    if(use_min){
      fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      make_fname(fname, min_hold.timestamp, min_hold.usec, MIN_HOLD);
      printf("Saving %s\n",fname);
      //stretch(min_hold.img_buffer, min_hold.width*min_hold.height, 0.01, 0.99);
      writeImage(fname, min_hold.width, min_hold.height, min_hold.img_buffer, "Finsprite minimum stack");
      free(fname);
      free(min_hold.img_buffer); min_hold.img_buffer = NULL;
    }

    if(use_min || use_dif){
      free(min_hold.img_buffer); min_hold.img_buffer = NULL;
    }
  }

}

static void errno_exit(const char *s){
  fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
  exit(EXIT_FAILURE);
}

static int xioctl(int fd, int request, void *arg){
  int r;
  
  do {
    r = ioctl(fd, request, arg);
  } while(r == -1 && EINTR == errno);
  
  return r;
}

char *make_time_string(time_t timestamp, long usec, int type){

  char *time_string = (char*)malloc(50*sizeof(char));

  strftime(time_string,30,"%Y-%m-%d_%H%M%S",gmtime(&(timestamp)));
  if(type == TRIG){
    char *tmp = (char*)malloc(50*sizeof(char));
    //    float f_usec = usec/1000000.0
    sprintf(tmp,".%.6ld",usec);
    strcat(time_string,tmp);
    free(tmp);
  }
  return time_string;
}

void make_fname(char *fname, time_t timestamp, long usec, int type){
  char tmp[50];
  mode_t dir_mode = S_IRWXU |S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  memset(fname,'\0',FNAME_SIZE);

  if(out_dir != NULL){ strcat(fname,out_dir); strcat(fname,"/"); }

  strftime(tmp,50,"%Y",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);

  if(mkdir(fname, dir_mode) < 0 && errno != EEXIST){
    printf("Unable to create %s\nSaving to current direcory: %d\n",fname,errno);
    // Empty path
    memset(fname,'\0',FNAME_SIZE);
  } else {
    strftime(tmp,50,"/%m",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    strftime(tmp,50,"/%d",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    strcat(fname,"/");
  }

  if(img_prefix != NULL){ strcat(fname,img_prefix); strcat(fname,"_"); }
  char *tmp2 = NULL; // malloc(30*sizeof(char));
  tmp2 = make_time_string(timestamp,usec,type);
  strcat(fname,tmp2);
  free(tmp2);

  switch(type){
  case AVE:
    sprintf(tmp,"_UTC_%d_s_ave.png", stack_length);
    break;
  case PEAK_HOLD:
    sprintf(tmp,"_UTC_%d_s_max.png", stack_length);
    break;
  case MIN_HOLD:
    sprintf(tmp,"_UTC_%d_s_min.png", stack_length);
    break;
  case DIF:
    sprintf(tmp,"_UTC_%d_s_dif.png", stack_length);
    break;
  case TRIG:
    sprintf(tmp,"_UTC_md.png");
    break;
  default:
    sprintf(tmp,".%ld_UTC.png",usec);
    break;
  }

  strcat(fname,tmp);

}

void save_ave_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  unsigned char *out_frame = calloc(ave_stack.width*ave_stack.height,sizeof(unsigned char));

  for(i=0;i<ave_stack.width*ave_stack.height;i++){
    //out_frame[i] = ave_stack.img_buffer[i]/ave_stack.num_imgs;
    //out_frame[i] = 255*ave_stack.img_buffer[i]/ave_stack.max_val;
    out_frame[i] = 255*(ave_stack.img_buffer[i]-ave_stack.min_val)/(ave_stack.max_val-ave_stack.min_val);
  }

  make_fname(fname, ave_stack.timestamp, ave_stack.usec, AVE);
  stretch(out_frame, ave_stack.width*ave_stack.height, 0.02, 0.98);
  printf("Saving %s, composed of %d frames\n",fname,ave_stack.num_imgs);
  writeImage(fname, ave_stack.width, ave_stack.height, out_frame, "Finsprite average stack");
  
  free(out_frame);
  free(fname);

  free(ave_stack.img_buffer);
  ave_stack.img_buffer = NULL;
  ave_stack.min_val = ULONG_MAX;
  ave_stack.max_val = 0;
  ave_stack.num_imgs = 0;
}

void save_dif_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  unsigned char *dif = calloc(min_hold.width*min_hold.height,sizeof(float));

  unsigned char *out_frame = calloc(min_hold.width*min_hold.height,sizeof(unsigned char));

  unsigned char dmin = UCHAR_MAX, dmax = 0;
  for(i=0;i<min_hold.width*min_hold.height;i++){
    dif[i] = peak_hold.img_buffer[i] - min_hold.img_buffer[i];
    if(dif[i] < dmin){ dmin = dif[i]; }
    if(dif[i] > dmax){ dmax = dif[i]; }
  }

  for(i=0;i<min_hold.width*min_hold.height;i++){
    out_frame[i] = 255*(dif[i]-dmin)/dmax;
  }
  
  free(dif); dif = NULL;

  make_fname(fname, min_hold.timestamp, min_hold.usec, DIF);
  printf("Saving %s\n",fname);
  //stretch(out_frame, min_hold.width*min_hold.height, 0.02, 0.98);
  writeImage(fname, min_hold.width, min_hold.height, out_frame, "Finsprite difference stack");
  
  free(out_frame); out_frame = NULL;
  free(fname); fname = NULL;

}

static int read_frame(void){
  struct v4l2_buffer buf;

  time_t frametime;

  CLEAR(buf);
    
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
    switch(errno) {
    case EAGAIN:
      printf("EAGAIN\n");
      return 0;
      
    case EIO:
      /* Could ignore EIO, see spec. */
      printf("EIO\n");
      /* fall through */
      
    default:
      printf("default\n");
      errno_exit("VIDIOC_DQBUF");
    }
  }
    
  assert(buf.index < n_mmap_buffers);
  
  frametime = buf.timestamp.tv_sec;

  if(!first_frame){
    process_image(mmap_buffers[buf.index].start, frametime, buf.timestamp.tv_usec);
  } else {
    first_frame = 0;
  }
  if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
    errno_exit("VIDIOC_QBUF");
    

  return 1;
}

static void mainloop(void){
  time_t timenow;
  time(&start_time);
  time(&timenow);
  while(difftime(timenow,start_time) < capture_length) {
    for(;;) {
      fd_set fds;
      struct timeval tv;
      int r;
      
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      /* Timeout. */
      tv.tv_sec = 2;
      tv.tv_usec = 0;
      
      r = select(fd + 1, &fds, NULL, NULL, &tv);
      
      if(r == -1) {
	if(EINTR == errno)
	  continue;
	
	errno_exit("select");
      }
      
      if(r == 0) {
	fprintf(stderr, "select timeout\n");
	exit(EXIT_FAILURE);
      }
      
      if(read_frame())
	break;
      
      /* EAGAIN - continue select loop. */
    }
    time(&timenow);
  }
}

static void stop_capturing(void){
  enum v4l2_buf_type type;
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
    errno_exit("VIDIOC_STREAMOFF");

}

static void start_capturing(void){
  unsigned int i;
  enum v4l2_buf_type type;
  
  for(i = 0; i < n_mmap_buffers; ++i) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMON, &type) == -1)
    errno_exit("VIDIOC_STREAMON");
    
}

static void uninit_device(void){
  unsigned int i;
  
  for(i = 0; i < n_mmap_buffers; ++i)
    if(munmap(mmap_buffers[i].start, mmap_buffers[i].length) == -1)
      errno_exit("munmap");
  
  free(mmap_buffers);
}

static void init_mmap(int n){
  struct v4l2_requestbuffers req;
  
  CLEAR(req);
  
  req.count = 2;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s does not support memory mapping\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_REQBUFS");
    }
  }
  
  if(req.count < 2) {
    fprintf(stderr, "Insufficient buffer memory on %s\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  mmap_buffers = calloc(req.count, sizeof(*mmap_buffers));
  
  if(!mmap_buffers) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }
  
  for(n_mmap_buffers = 0; n_mmap_buffers < req.count; ++n_mmap_buffers) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n_mmap_buffers;
    
    if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
      errno_exit("VIDIOC_QUERYBUF");
    
    mmap_buffers[n_mmap_buffers].length = buf.length;
    mmap_buffers[n_mmap_buffers].start =
      mmap(NULL /* start anywhere */,
	   buf.length,
	   PROT_READ | PROT_WRITE /* required */,
	   MAP_SHARED /* recommended */,
	   fd, buf.m.offset);
    
    if(MAP_FAILED == mmap_buffers[n_mmap_buffers].start)
      errno_exit("mmap");
  }
}

static void init_device(int img_in_mmap_buffer){
  struct v4l2_capability cap;
  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;
  unsigned int min;
  
  if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s is no V4L2 device\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_QUERYCAP");
    }
  }
  
  if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf(stderr, "%s is no video capture device\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf(stderr, "%s does not support streaming i/o\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  /* Select video input, video standard and tune here. */
  
  CLEAR(cropcap);

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c = cropcap.defrect; /* reset to default */
    
    if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
      switch(errno) {
      case EINVAL:
	/* Cropping not supported. */
	break;
      default:
	/* Errors ignored. */
	break;
      }
    }
  } else {
    /* Errors ignored. */
  }
    
  CLEAR(fmt);
  
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = 720;
  fmt.fmt.pix.height = 576;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

  if(xioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
    errno_exit("VIDIOC_S_FMT");
  
  /* Note VIDIOC_S_FMT may change width and height. */
  
  /* Buggy driver paranoia. */
  min = fmt.fmt.pix.width * 2;
  if(fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if(fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  init_mmap(img_in_mmap_buffer);

}

static void close_device(void){
  if(close(fd) == -1)
    errno_exit("close");
  
  fd = -1;
}

static void open_device(void){
  struct stat st; 
  
  if(stat(dev_name, &st) == -1) {
    fprintf(stderr, "Cannot identify '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  if(!S_ISCHR(st.st_mode)) {
    fprintf(stderr, "%s is no device\n", dev_name);
    exit(EXIT_FAILURE);
  }
  
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  
  if(fd == -1) {
    fprintf(stderr, "Cannot open '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
}

static void usage (FILE *fp, int argc, char **argv){
  fprintf(fp,
	  "Usage: %s [options]\n\n"
	  "Options:\n"
	  "-c | --capture-device <name> Video device name [/dev/video0]\n"
	  "-h | --help  Print this message\n"
	  "-n | --buffer <n>  Number of images in ring buffer [10]\n"
	  "                   Enables motion detection\n"
	  "-s | --stack-length <s>  Length of stacked period in seconds [60]\n"
	  "-m | --min  Enable minimum stacking [default: off]\n"
	  "-M | --max  Enable maximum stacking [default: off]\n"
	  "-a | --ave  Enable average stacking [default: off]\n"
	  "-d | --dif  Enable difference stacking [default: off]\n"
	  "-t | --trig  Enable motion detection with default ring buffer [default: off]\n"
	  "-T | --trigger-threshold  Threshold for motion detection [default: 1000]\n"
	  "-x | --x-regions  Number of motion detecion zones in horizontal direction [4]\n"
	  "-y | --y-regions  Number of motion detecion zones in vertical direction [3]\n"
	  "-k | --mask  Mask filename used to exclude areas from motion detect\n"
	  "-G | --trigger-test  Test how sensitive triggering is. No images are saved\n"
	  "-l | --capture-length <time>  Length of imaging period in seconds [unlimited]\n"
	  "-p | --prefix <prefix>  Image filename prefix [default: none]\n"
	  "-o | --out-dir <dir>  Image output directory [default: current directory]\n"
	  "                      The images are saved to <dir>/yyyy/mm/dd/\n"
	  "\n",
	  argv[0]);
}

int main(int argc, char **argv){
  dev_name = "/dev/video0";

  for(;;) {
    int index;
    int c;
    
    c = getopt_long(argc, argv,
		    short_options, long_options,
		    &index);
    
    if(c == -1)
      break;
    
    switch(c) {
    case 0: /* getopt_long() flag */
      break;
    case 'c':
      dev_name = optarg;
      break;      
    case 'h':
      usage(stdout, argc, argv);
      exit(EXIT_SUCCESS);
    case 'n':
      sscanf(optarg,"%d",&img_in_md_buffer);
      use_trig = 1;
      break;
    case 's':  /* length of stacked period in secs */
      sscanf(optarg,"%d",&stack_length);
      break;
    case 'm': /* Enable minimum stack */
      use_min = 1;
      break;
    case 'M': /* Enable maximum stack */
      use_max = 1;
      break;
    case 'a': /* Enable average stack */
      use_ave = 1;
      break;
    case 't': /* Enable triggering (motion detection)*/
      use_trig = 1;
      break;
    case 'T': /* Trigger/motion detect threshold*/
      printf("%s\n",optarg);
      sscanf(optarg,"%d",&md_threshold);
      break;
    case 'x': /* Number of triggering regions in x-direction */
      sscanf(optarg,"%d",&md_num_x_zones);
      break;
    case 'y': /* Number of triggering regions in y-direction */
      sscanf(optarg,"%d",&md_num_y_zones);
      break;
    case 'k': /* Mask that can be used to exclude pixels from motion detection */
      md_mask_fname = optarg;
      break;
    case 'G':
      md_test = 1;
      use_trig = 1;
      break;
    case 'P':
      sscanf(optarg,"%d",&md_percentage);
      break;
    case 'd': /* Enable difference stack */
      use_dif = 1;
      break;
    case 'l': /* Length of imaging period */
      sscanf(optarg,"%ld",&capture_length);
      break;
    case 'p': /* Image prefix */
      img_prefix = optarg;
      break;
    case 'o': /* Output directory prefix */
      out_dir = optarg;
      break;
    default:
      usage(stderr, argc, argv);
      exit(EXIT_FAILURE);
    }
  }

  if(!use_min && !use_max && !use_ave && !use_trig && !use_dif){
    printf("\nNothing to do, terminating.\n");
    printf("Specify at least one imaging mode!\n\n");
    usage(stderr, argc, argv);
    exit(EXIT_FAILURE);
  }

  if(md_test){
    use_min = 0;
    use_max = 0;
    use_ave = 0;
    use_dif = 0;
  }

  int i;
  ave_stack.img_buffer = NULL;
  ave_stack.max_val = 0;
  ave_stack.max_val = ULONG_MAX;
  
  peak_hold.img_buffer = NULL;
  min_hold.img_buffer = NULL;

  if(use_trig){
    md_frames = calloc(img_in_md_buffer,sizeof(*md_frames));
    for(i=0; i<img_in_md_buffer; i++){
      md_frames[i].img_buffer = NULL;
      md_frames[i].width = -1;
      md_frames[i].height = -1;
    }
  }

  open_device();

  init_device(img_in_mmap_buffer);

  start_capturing();

  mainloop();

  stop_capturing();

  uninit_device();

  close_device();

  if(use_trig){
    for(i=0; i < img_in_md_buffer; i++){
      if(md_frames[i].img_buffer != NULL){
	free(md_frames[i].img_buffer);
	md_frames[i].img_buffer = NULL;
      }
    }
    if(md_zone_sums != NULL){ free(md_zone_sums); }
    if(md_zone_sums_new != NULL){ free(md_zone_sums_new); }
    if(md_zone_num_pix != NULL){ free(md_zone_num_pix); }
    if(md_zone_reference != NULL){ free(md_zone_reference); }
  }

  if(md_frames != NULL){ free(md_frames); }
  if(peak_hold.img_buffer != NULL){ free(peak_hold.img_buffer); }
  if(min_hold.img_buffer != NULL){ free(min_hold.img_buffer); }
  if(ave_stack.img_buffer != NULL){ free(ave_stack.img_buffer); }

  exit(EXIT_SUCCESS);

  return 0;
}
